from collections import UserDict


class GroupItem(UserDict):
    """Group item."""

    class Scripts:
        class Selectors:
            url = 'a.groups__avatar'
        url = 'n => n.getAttribute("href")'

    s = Scripts
    ss = Scripts.Selectors

    def __init__(self, initialdata):
        super().__init__(initialdata)

    def __repr__(self):
        return repr(self.data)

    @classmethod
    async def from_element(cls, element):
        """Creates a new group item from a DOM element.

        Args:
            element (pyppeteer.element_handle.ElementHandle): the element.

        Returns:
            item (GroupItem): new instance of this class.

        """

        url = await element.Jeval(cls.ss.url, cls.s.url)
        data = {'link': url.strip('?ref=')}
        group = cls(initialdata=data)
        return group



import http.cookies
import re
from datetime import datetime
from enum import Enum

EMAIL_PATTERN = r"(^[a-zA-Z0-9_.+-]+)@([a-zA-Z0-9-]+)\.([a-zA-Z0-9-.]+$)"
PRIVILEGES = ['photos', 'guestbook', 'stream', 'messages', 'events']


def full_scope():
    return ' '.join(PRIVILEGES)


class SignatureCircuit(Enum):
    """Signature circuit.

    .. _Подпись запроса
        https://api.mail.ru/docs/guides/restapi/#sig

    """

    UNDEFINED = 0
    CLIENT_SERVER = 1
    SERVER_SERVER = 2


def parseaddr(address):
    """Converts an e-mail address to a tuple - (screen name, domain name).

    Args:
        address (str): e-mail address

    Returns:
        domain_name(str): domain name
        screen_name (str): screen name

    """

    pattern = re.compile(EMAIL_PATTERN)
    match = pattern.match(address)

    if match is None:
        raise ValueError("email address %r is not valid" % address)

    screen_name, domain_name, _ = match.groups()
    return domain_name, screen_name


class Cookie(dict):
    """Represents cookie in a browser."""

    expires_fmt = '%a, %d %b %Y %H:%M:%S GMT'

    def __init__(self, *args):
        super().__init__(*args)

    @classmethod
    def from_morsel(cls, morsel):
        """Converts a cookie morsel to dictionary.

        Args:
            morsel (http.cookies.Morsel): cookie morsel

        Returns:
            cookie (dict): cookie for the browser.

        """

        domain = morsel['domain']
        expires = morsel['expires']
        path = morsel['path']
        size = len(morsel.key) + len(morsel.value)
        http_only = True if morsel['httponly'] else False
        secure = True if morsel['secure'] else False

        if expires:
            session = False
            expires = datetime.strptime(expires, cls.expires_fmt).timestamp()
        else:
            session = True
            expires = None

        if not domain.startswith('.'):
            domain = '.' + domain

        cookie = cls({
            'name': morsel.key,
            'value': morsel.value,
            'domain': domain,
            'path': path,
            'expires': expires,
            'size': size,
            'httpOnly': http_only,
            'secure': secure,
            'session': session,
        })

        return cookie

    @classmethod
    def to_morsel(cls, cookie):
        """Converts a dictionary to cookie morsel.

        Args:
            cookie (dict): cookie from the browser.

        Returns:
            morsel (http.cookies.Morsel): cookie morsel

        """

        morsel = http.cookies.Morsel()
        morsel.set(cookie['name'], cookie['value'], cookie['value'])

        if cookie['expires']:
            dt = datetime.fromtimestamp(cookie['expires'])
            morsel['expires'] = dt.strftime(cls.expires_fmt)
        else:
            dt = datetime.fromtimestamp(0.)
            morsel['expires'] = dt.strftime(cls.expires_fmt)
        if cookie['path']:
            morsel['path'] = cookie['path']
        if cookie['domain']:
            morsel['domain'] = cookie['domain']
        if cookie['secure']:
            morsel['secure'] = cookie['secure']
        if cookie['httpOnly']:
            morsel['httponly'] = cookie['httpOnly']

        return morsel



"""My.Mail.Ru API."""

from .sessions import TokenSession


class API:
    """Platform@Mail.Ru REST API."""

    __slots__ = ('session', )

    def __init__(self, session: TokenSession):
        self.session = session

    def __getattr__(self, name):
        return APIMethod(self, name)

    async def __call__(self, name, **params):
        return await getattr(self, name)(**params)


class APIMethod:
    """Platform@Mail.Ru REST API method."""

    __slots__ = ('api', 'name')

    def __init__(self, api: API, name: str):
        self.api = api
        self.name = name

    def __getattr__(self, name):
        return APIMethod(self.api, f'{self.name}.{name}')

    async def __call__(self, **params):
        params['method'] = self.name
        return await self.api.session.request(params=params)



import logging
import os
from pyppeteer import connect, launch


log = logging.getLogger(__name__)


class Browser:
    """A wrapper around pyppeteer.browser.Browser."""

    endpoint = os.environ.get('PYPPETEER_BROWSER_ENDPOINT')
    viewport = os.environ.get('PYPPETEER_BROWSER_VIEWPORT', '800,600')
    slow_mo = int(os.environ.get('PYPPETEER_BROWSER_SLOW_MO', '200'))

    def __init__(self, browser=None):
        self.browser = browser
        self.contexts = {}

    def __await__(self):
        return self.start().__await__()

    async def start(self):
        """Starts chrome process or connects to the existing chrome."""

        if self.browser:
            pass
        elif self.endpoint:
            browser_conn = {'browserWSEndpoint': self.endpoint}
            log.debug('connecting: {}'.format(browser_conn))
            self.browser = await connect(browser_conn, slowmMo=self.slow_mo)
        else:
            log.debug('launching new browser..')
            self.browser = await launch(slowMo=self.slow_mo)

        return self

    async def page(self, url, session_key,
                   cookies=(), force=False, context=None):
        """Makes new page and returns its object.

        Args:
            url (str): URL to navigate page to. The url should
                include scheme, e.g. `https://`.
            session_key (str): access token.
            cookies (tuple): cookies for the page.
            force (bool): `True` - to always return a new context.
            context (pyppeteer.browser.BrowserContext): browser context.

        Returns:
            page (pyppeteer.page.Page): page.

        """

        if not self.browser:
            await self.start()

        if context:
            pass
        elif (url, session_key) in self.contexts:
            context = self.contexts[(url, session_key)]
        elif force:
            context = await self.browser.createIncognitoBrowserContext()
        else:
            context = await self.browser.createIncognitoBrowserContext()
            self.contexts[(url, session_key)] = context

        blank_page = None
        for target in context.targets():
            if target.url == 'about:blank':
                blank_page = await target.page()
            elif target.url == url:
                page = await target.page()
                break
        else:
            page = blank_page or await context.newPage()
            viewport = ('width', 'height'), map(int, self.viewport.split(','))
            await page.setViewport(dict(zip(*viewport)))
            await page.setCookie(*cookies)
            await page.setRequestInterception(True)

            @page.on('request')
            async def on_request(request):
                if request.url.endswith('.png') or request.url.endswith('.jpg'):
                    await request.abort()
                else:
                    await request.continue_()

            await page.goto(url)

        return page



class Error(Exception):

    @property
    def error(self):
        return self.args[0]

    def __init__(self, error: str or dict):
        arg = error if isinstance(error, dict) else {
            'error': 'internal_error',
            'error_description': error,
        }
        super().__init__(arg)


class OAuthError(Error):
    """OAuth error."""

    def __init__(self, error: str):
        super().__init__({'error': 'oauth_error', 'error_description': error})


class CustomOAuthError(Error):
    """Custom errors that raised when authorization failed."""

    ERROR = {'error': '', 'error_description': ''}

    def __init__(self):
        super().__init__(self.ERROR)


class InvalidGrantError(CustomOAuthError):
    """Invalid user credentials."""

    ERROR = {
        'error': 'invalid_grant',
        'error_description': 'invalid login or password',
    }


class InvalidClientError(CustomOAuthError):
    """Invalid client id."""

    ERROR = {
        'error': 'invalid_client',
        'error_description': 'invalid client id',
    }


class InvalidUserError(CustomOAuthError):
    """Invalid user (blocked)."""

    ERROR = {
        'error': 'invalid_user',
        'error_description': 'user is blocked',
    }


class APIError(Error):
    def __init__(self, error):
        super().__init__(error)
        self.code = error['error_code']
        self.msg = error['error_msg']

    def __str__(self):
        return f'Error {self.code}: {self.msg}'


class APIScrapperError(Error):
    code = 0

    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return f'Error {self.code}: {self.msg}'


class CustomAPIError(Error):
    ERROR = {'error_code': 0, 'error_msg': ''}

    def __init__(self):
        super().__init__(self.ERROR)


class EmptyObjectsError(CustomAPIError):
    ERROR = {'error_code': 202, 'error_msg': 'empty objects'}


class EmptyGroupsError(CustomAPIError):
    ERROR = {'error_code': 202, 'error_msg': 'empty groups'}


class AccessDeniedError(CustomAPIError):
    ERROR = {
        'error_code': 202,
        'error_msg': 'Access to this object is denied',
    }


class BlackListError(CustomAPIError):
    ERROR = {
        'error_code': 202,
        'error_msg': 'Access to this object is denied: you are in blacklist',
    }


class CookieError(APIScrapperError):
    code = 1



from collections import UserDict


class Event(UserDict):
    """Event."""

    class Scripts:
        class Selectors:
            event_class = 'b-history_event_active-area_shift'
            subevent_class = 'b-history_event_active-area'
            event = f'div.{event_class}'
            subevent = f'div.{subevent_class}:not(.{event_class})'

            head = 'div.b-history-event_head'
            action = f'{head} div.b-history-event__action'
            author = f'{action} .b-history-event__ownername'
            time = f'{action} div.b-history-event_time'
            url = f'{time} a'

            text = (
                'div.b-history-event__body '
                'div.b-history-event__event-textbox2 '
            )
            status = (
                'div.b-history-event__body '
                'div.b-history-event__event-textbox_status '
            )
            links = f'{status} a'

            comments = 'div.b-comments__history'

        astat = 'n => n.getAttribute("data-astat")'
        author = 'n => n.getAttribute("href")'
        url = 'n => n.getAttribute("href")'

        links = (
            'ns => ns.map(n => {'
            'return { href: n.getAttribute("href"), text: n.innerText };'
            '})'
        )
        text = 'n => n.innerText'
        status = 'n => n.innerText'

    class Types:
        clickable = ['1-1', '3-23', '5-39', '5-41']
        status = '3-23'

    s = Scripts
    ss = Scripts.Selectors
    t = Types

    def __init__(self, initialdata):
        super().__init__(initialdata)

    def __repr__(self):
        return repr(self.data)

    @classmethod
    async def from_element(cls, element):
        """Creates a new event from a DOM element.

        Args:
            element (pyppeteer.element_handle.ElementHandle): the element.

        Returns:
            event (Event): new instance of this class.

        """

        ctx = element.executionContext
        astat = await ctx.evaluate(cls.s.astat, element)
        astat = Astat(*astat.split(':'))
        comments = await element.J(cls.ss.comments)

        if astat.subtype in ['comment', 'like']:
            data = {
                'time': astat.time,
                'author': {},  # fixed below
                # TODO: scrape like/comment 'huid'
                'subevent': {
                    # TODO: scrape subevent 'thread_id'
                    'authors': [],  # fixed below
                    'type_name': astat.corr_type_name,
                    # skip 'click_url', added below if present
                    'likes_count': astat.likes_count,
                    # skip 'attachments', added below if present
                    # TODO: scrape subevent 'time'
                    # TODO: scrape subevent 'huid'
                    # TODO: scrape subevent 'generator'
                    'user_text': '',  # fixed below if present
                    # TODO: scrape subevent 'is_liked_by_me'
                    'subtype': 'event',
                    'is_commentable': 1 if comments else 0,
                    'type': astat.corr_type,
                    'is_likeable': 1 if comments else 0,
                    'id': astat.corr_event_id,
                    # skip 'text_media', added below if present
                    'comments_count': astat.comments_count,
                    # TODO: scrape subevent 'action_links'
                },
                'subtype': astat.subtype,
                'is_commentable': 0,
                'id': astat.id,
                'is_likeable': 0,
            }

            element = await element.J(cls.ss.subevent)
            element_type = astat.corr_type
            element_body = await cls._from_element(element, element_type)
            data['subevent'].update(element_body)

        else:
            data = {
                # TODO: scrape event 'thread_id'
                'authors': [],  # fixed below
                'type_name': astat.type_name,
                # skip 'click_url', added below if present
                'likes_count': astat.likes_count,
                # skip 'attachments', added below if present
                'time': astat.time,
                # TODO: scrape event 'huid'
                # TODO: scape event 'generator'
                'user_text': '',  # fixed below if present
                # TODO: scrape event 'is_liked_by_me'
                'subtype': astat.subtype,
                'is_commentable': 1 if comments else 0,
                'type': astat.type,
                'is_likeable': 1 if comments else 0,
                'id': astat.id,
                # skip 'text_media', added below if present
                'comments_count': astat.comments_count,
                # TODO: scrape event 'action_links'
            }

            element = await element.J(cls.ss.event)
            element_type = astat.type
            element_body = await cls._from_element(element, element_type)
            data.update(element_body)

        event = cls(initialdata=data)

        return event

    @classmethod
    async def _from_element(cls, elem, elem_type):
        body, author, url = {}, {}, ''

        # scrape 'authors' (all but the stream owner)
        author_ref = await elem.Jeval(cls.ss.author, cls.s.author) or ''
        if author_ref:
            author['link'] = author_ref.rstrip('?ref=ho')
        body['authors'] = [author] if author else []

        # scrape 'click_url'
        if elem_type in cls.t.clickable:
            url = await elem.Jeval(cls.ss.url, cls.s.url)
        body['click_url'] = f'https://my.mail.ru{url}' if url else url

        # scrape 'user_text' and 'text_media'
        if elem_type == cls.t.status:
            stts = await elem.J(cls.ss.status)
            text = await elem.Jeval(cls.ss.status, cls.s.status) if stts else ''
            links = await stts.JJeval(cls.ss.links, cls.s.links)

            for link in links:
                text = text.replace(link['text'], link['href'])

            # scrape 'text_media'
            content = {'type-id': 'text', 'contents': text}
            link_media = [{'object': 'link', 'content': content} for _ in links]
            text_media = [{'object': 'text', 'content': text}]
            body['text_media'] = link_media + text_media
        else:
            text = await elem.J(cls.ss.text)
            text = await elem.Jeval(cls.ss.text, cls.s.text) if text else ''
        body['user_text'] = text

        return body


class Astat:
    def __init__(self, user_world_id, event_type, event_id,
                 owner_world_id, corr_world_id, corr_event_id,
                 likes_count, comments_count, event_time, *_):
        self.user_world_id = int(user_world_id or '0')

        self.event_type = event_type
        self.event_id = event_id
        self.event_time = int(event_time)
        self.owner_world_id = owner_world_id

        self.corr_world_id = corr_world_id
        self.corr_event_id = corr_event_id

        self.likes_count = int(likes_count or '0')
        self.comments_count = int(comments_count or '0')

    @property
    def id(self):
        return self.event_id.lower()

    @property
    def time(self):
        return self.event_time

    @property
    def type(self):
        """Event type."""
        if self.subtype == 'event':
            return '-'.join(self.event_type.split('-')[:2])
        else:
            raise AttributeError()

    @property
    def corr_type(self):
        """Type of liked/commented subevent."""
        if self.subtype == 'event':
            raise AttributeError()
        else:
            return '-'.join(self.event_type.split('-')[:2])

    @property
    def type_name(self):
        return TYPE_NAMES.get(self.type, '')

    @property
    def corr_type_name(self):
        return TYPE_NAMES[self.corr_type]

    @property
    def subtype(self):
        type_code = self.event_type.split('-')
        if len(type_code) < 3:
            return 'event'
        else:
            return type_code[2].lower()


TYPE_NAMES = {
    '1-1': 'photo_upload',
    '1-2': 'video_upload',
    '1-7': 'music_add',
    '3-3': 'user_community_actions_enter',
    '3-5': 'user_community_actions_leave',
    '3-23': 'micropost',
    '5-7': 'avatar_change',
    '5-10': 'gift_send',
    '5-11': 'gift_received',
    '5-16': 'app_add',
    '5-26': 'share',
    '5-28': 'app_info2',
    '5-37': 'gift_receive_multi',
    '5-39': 'community_post',
    '5-41': 'user_post',
    '5-44': 'community_video_upload',
    '5-47': 'community_photo_upload',
    '5-50': '',  # TODO: add name
    #  TODO: add missing types
}



import aiohttp
import asyncio
import hashlib
import logging
from yarl import URL

from .exceptions import (
    Error,
    OAuthError,
    InvalidGrantError,
    InvalidClientError,
    InvalidUserError,
    APIError,
)
from .parser import AuthPageParser
from .utils import full_scope, parseaddr, SignatureCircuit, Cookie


log = logging.getLogger(__name__)


class Session:
    """A wrapper around aiohttp.ClientSession."""

    __slots__ = ('pass_error', 'session')

    def __init__(self, pass_error=False, session=None):
        self.pass_error = pass_error
        self.session = session or aiohttp.ClientSession()

    def __await__(self):
        return self.authorize().__await__()

    async def __aenter__(self):
        return await self.authorize()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def authorize(self):
        return self

    async def close(self):
        await self.session.close()


class PublicSession(Session):
    """Session for calling public API methods of Platform@Mail.Ru."""

    PUBLIC_URL = 'http://appsmail.ru/platform'
    CONTENT_TYPE = 'text/javascript; charset=utf-8'

    async def public_request(self, segments=(), params=None):
        """Requests public data.

        Args:
            segments (tuple): additional segments for URL path.
            params (dict): URL parameters.

        Returns:
            response (dict): JSON object response.

        """

        url = f'{self.PUBLIC_URL}/{"/".join(segments)}'

        try:
            async with self.session.get(url, params=params) as resp:
                content = await resp.json(content_type=self.CONTENT_TYPE)
        except aiohttp.ContentTypeError:
            msg = f'got non-REST path: {url}'
            log.error(msg)
            raise Error(msg)

        if self.pass_error:
            response = content
        elif 'error' in content:
            log.error(content['error'])
            raise APIError(content['error'])
        else:
            response = content

        return response


class TokenSession(PublicSession):
    """Session for sending authorized requests."""

    API_URL = f'{PublicSession.PUBLIC_URL}/api'
    ERROR_MSG = 'See https://api.mail.ru/docs/guides/restapi/#sig.'

    __slots__ = ('app_id', 'private_key', 'secret_key', 'session_key', 'uid')

    def __init__(self, app_id, private_key, secret_key, access_token, uid,
                 cookies=(), pass_error=False, session=None):
        super().__init__(pass_error, session)
        self.app_id = app_id
        self.private_key = private_key
        self.secret_key = secret_key
        self.session_key = access_token
        self.uid = uid
        self.cookies = cookies

    @property
    def cookies(self):
        """HTTP cookies from cookie jar."""
        return [Cookie.from_morsel(m) for m in self.session.cookie_jar]

    @cookies.setter
    def cookies(self, cookies):
        loose_cookies = []

        for cookie in cookies:
            loose_cookie = Cookie.to_morsel(cookie)
            loose_cookies.append((loose_cookie.key, loose_cookie))

        self.session.cookie_jar.update_cookies(loose_cookies)

    @property
    def sig_circuit(self):
        if self.uid and self.private_key:
            return SignatureCircuit.CLIENT_SERVER
        elif self.secret_key:
            return SignatureCircuit.SERVER_SERVER
        else:
            return SignatureCircuit.UNDEFINED

    @property
    def required_params(self):
        """Required parameters."""
        params = {'app_id': self.app_id, 'session_key': self.session_key}
        if self.sig_circuit is SignatureCircuit.SERVER_SERVER:
            params['secure'] = '1'
        return params

    def params_to_str(self, params):
        query = ''.join(f'{k}={str(params[k])}' for k in sorted(params))

        if self.sig_circuit is SignatureCircuit.CLIENT_SERVER:
            return f'{self.uid}{query}{self.private_key}'
        elif self.sig_circuit is SignatureCircuit.SERVER_SERVER:
            return f'{query}{self.secret_key}'
        else:
            raise Error(self.ERROR_MSG)

    def sign_params(self, params):
        """Signs the request parameters according to signature circuit.

        Args:
            params (dict): request parameters

        Returns:
            sig (str): signature

        """

        query = self.params_to_str(params)
        sig = hashlib.md5(query.encode('UTF-8')).hexdigest()
        return sig

    async def request(self, segments=(), params=()):
        """Sends a request.

        Args:
            segments (tuple): additional segments for URL path.
            params (dict): URL parameters, contains key 'method', e.g.
                {
                    "method": "stream.getByAuthor",
                    "uid": "15410773191172635989",
                    "limit": 10,
                }

        Returns:
            response (dict): JSON object response.

        """

        url = f'{self.API_URL}/{"/".join(segments)}'

        params = {k: params[k] for k in params if params[k]}
        params.update(self.required_params)
        params.update({'sig': self.sign_params(params)})

        async with self.session.get(url, params=params) as resp:
            content = await resp.json(content_type=self.CONTENT_TYPE)

        if self.pass_error:
            response = content
        elif 'error' in content:
            log.error(content['error'])
            raise APIError(content['error'])
        else:
            response = content

        return response


class ClientSession(TokenSession):

    ERROR_MSG = 'Pass "uid" and "private_key" to use client-server circuit.'

    def __init__(self, app_id, private_key, access_token, uid, cookies=(),
                 pass_error=False, session=None):
        super().__init__(app_id, private_key, '', access_token, uid, cookies,
                         pass_error, session)


class ServerSession(TokenSession):

    ERROR_MSG = 'Pass "secret_key" to use server-server circuit.'

    def __init__(self, app_id, secret_key, access_token, cookies=(),
                 pass_error=False, session=None):
        super().__init__(app_id, '', secret_key, access_token, '', cookies,
                         pass_error, session)


class ImplicitSession(TokenSession):

    OAUTH_URL = 'https://connect.mail.ru/oauth/authorize'
    REDIRECT_URI = 'http%3A%2F%2Fconnect.mail.ru%2Foauth%2Fsuccess.html'

    AUTHORIZE_NUM_ATTEMPTS = 1
    AUTHORIZE_RETRY_INTERVAL = 3

    GET_AUTH_DIALOG_ERROR_MSG = 'Failed to open authorization dialog.'
    POST_AUTH_DIALOG_ERROR_MSG = 'Form submission failed.'
    GET_ACCESS_TOKEN_ERROR_MSG = 'Failed to receive access token.'

    __slots__ = ('email', 'passwd', 'scope',
                 'expires_in', 'refresh_token', 'token_type')

    def __init__(self, app_id, private_key, secret_key, email, passwd, scope,
                 pass_error=False, session=None):
        super().__init__(app_id, private_key, secret_key, '', '', (),
                         pass_error, session)
        self.email = email
        self.passwd = passwd
        self.scope = scope or full_scope()

    @property
    def params(self):
        """Authorization parameters."""
        return {
            'client_id': self.app_id,
            'redirect_uri': self.REDIRECT_URI,
            'response_type': 'token',
            'scope': self.scope,
        }

    async def authorize(self, num_attempts=None, retry_interval=None):
        """OAuth Implicit flow."""

        num_attempts = num_attempts or self.AUTHORIZE_NUM_ATTEMPTS
        retry_interval = retry_interval or self.AUTHORIZE_RETRY_INTERVAL

        for attempt_num in range(num_attempts):
            log.debug(f'getting authorization dialog {self.OAUTH_URL}')
            url, html = await self._get_auth_dialog()

            if 'Не указано приложение' in html:
                raise InvalidClientError()
            elif url.path == '/oauth/authorize':
                log.debug(f'authorizing at {url}')
                url, html = await self._post_auth_dialog(html)

            if url.path == '/oauth/success.html':
                await self._get_access_token()
                return self
            elif url.path == '/recovery':
                log.error(f'User {self.email} is blocked.')
                raise InvalidUserError()
            elif url.query.get('fail') == '1':
                log.error(f'Invalid e-mail {self.email} or password.')
                raise InvalidGrantError()

            await asyncio.sleep(retry_interval)
        else:
            log.error(f'{num_attempts} login attempts exceeded.')
            raise OAuthError(f'{num_attempts} login attempts exceeded.')

    async def _get_auth_dialog(self):
        """Returns url and html code of authorization dialog."""

        async with self.session.get(self.OAUTH_URL, params=self.params) as resp:
            if resp.status != 200:
                log.error(self.GET_AUTH_DIALOG_ERROR_MSG)
                raise OAuthError(self.GET_AUTH_DIALOG_ERROR_MSG)
            else:
                url, html = resp.url, await resp.text()

        return url, html

    async def _post_auth_dialog(self, html):
        """Submits a form with e-mail, password and domain
        to get access token and user id.

        Args:
            html (str): authorization page's html code.

        Returns:
            url (URL): redirected page's url.
            html (str): redirected page's html code.

        """

        parser = AuthPageParser()
        parser.feed(html)
        form_url, form_data = parser.form
        parser.close()

        domain, login = parseaddr(self.email)
        form_data['Login'] = login
        form_data['Domain'] = f'{domain}.ru'
        form_data['Password'] = self.passwd

        async with self.session.post(form_url, data=form_data) as resp:
            if resp.status != 200:
                log.error(self.POST_AUTH_DIALOG_ERROR_MSG)
                raise OAuthError(self.POST_AUTH_DIALOG_ERROR_MSG)
            else:
                url, html = resp.url, await resp.text()

        return url, html

    async def _get_access_token(self):
        async with self.session.get(self.OAUTH_URL, params=self.params) as resp:
            if resp.status != 200:
                log.error(self.GET_ACCESS_TOKEN_ERROR_MSG)
                raise OAuthError(self.GET_ACCESS_TOKEN_ERROR_MSG)
            else:
                location = URL(resp.history[-1].headers['Location'])
                url = URL(f'?{location.fragment}')

        try:
            self.session_key = url.query['access_token']
            self.expires_in = url.query['expires_in']
            self.refresh_token = url.query['refresh_token']
            self.token_type = url.query['token_type']
            self.uid = url.query['x_mailru_vid']
        except KeyError as e:
            raise OAuthError(f'"{e.args[0]}" is missing in the auth response')


class ImplicitClientSession(ImplicitSession):
    def __init__(self, app_id, private_key, email, passwd, scope,
                 pass_error=False, session=None):
        super().__init__(app_id, private_key, '', email, passwd, scope,
                         pass_error, session)


class ImplicitServerSession(ImplicitSession):
    def __init__(self, app_id, secret_key, email, passwd, scope,
                 pass_error=False, session=None):
        super().__init__(app_id, '', secret_key, email, passwd, scope,
                         pass_error, session)



"""My.Mail.Ru scrapers."""

import asyncio
import logging
from functools import wraps

from .exceptions import (
    APIError,
    APIScrapperError,
    CookieError,
    EmptyObjectsError,
    EmptyGroupsError,
    AccessDeniedError,
    BlackListError,
)
from .api import API, APIMethod
from .browser import Browser
from .objects import Event, GroupItem
from .sessions import TokenSession


log = logging.getLogger(__name__)


class APIScraper(API, Browser):
    """API scraper."""

    __slots__ = ('browser', )

    def __init__(self, session: TokenSession, browser=None):
        API.__init__(self, session)
        Browser.__init__(self, browser)

    def __getattr__(self, name):
        return scrapers.get(name, APIMethod)(self, name)


class APIScraperMethod(APIMethod):
    """API scraper's method."""

    class Scripts:
        """Common scripts."""
        class Selectors:
            """Common selectors."""
            content = (
                ' html body'
                ' div.l-content'
                ' div.l-content__center'
                ' div.l-content__center__inner'
            )
            main_page = f'{content} div.b-community__main-page'
            closed_signage = f'{main_page} div.mf_cc'
            profile = f'{main_page} div.profile'
            profile_content = f'{profile} div.profile__contentBlock'

            class SelectorTemplates:
                """Common templates of selectors."""
                hidden = '%s[style="display: none;"]'
                visible = '%s:not([style="display: none;"])'

        class ScriptTemplates:
            """Common templates of scripts."""
            getattr = 'n => n.getAttribute("%s")'
            getclass = getattr % 'class'
            getstyle = getattr % 'style'
            selector = 'document.querySelector("%s")'
            selector_all = 'document.querySelectorAll("%s")'
            click = f'{selector}.click()'
            computed_style = f'window.getComputedStyle({selector})'
            display = f'{computed_style}["display"]'
            visible = f'{display} != "none"'
            length = f'{selector_all}.length'

        scroll = 'window.scroll(0, document.body.scrollHeight)'

    s = Scripts
    ss = Scripts.Selectors
    sst = Scripts.ScriptTemplates
    ssst = Scripts.Selectors.SelectorTemplates

    def __init__(self, api: APIScraper, name: str):
        super().__init__(api, name)
        self.page = None

    def __getattr__(self, name):
        name = f'{self.name}.{name}'
        return scrapers.get(name, APIMethod)(self.api, name)

    async def __call__(self, scrape=False, **params):
        call = self.call if scrape else super().__call__
        return await call(**params)

    async def init(self, **params):
        pass

    async def call(self, **params):
        await self.init(**params)


class APIScraperMultiMethod(APIScraperMethod):

    multiarg = 'uids'
    empty_objects_error = EmptyObjectsError
    ignored_errors = (APIError, )

    async def __call__(self, scrape=False, **params):
        call = self.multicall if scrape else super().__call__()
        return await call(**params)

    async def multicall(self, **params):
        args = params[self.multiarg].split(',')
        result = []
        for arg in args:
            params.pop(self.multiarg)
            params.update({self.multiarg: arg})

            # when `self.api.session.pass_error` is False
            try:
                resp = await self.call(**params)
            except self.ignored_errors:
                resp = self.empty_objects_error().error

            # when `self.api.session.pass_error` is True
            if isinstance(resp, dict) and 'error_code' in resp:
                pass
            else:
                result.append(resp[0])

        if not result and self.empty_objects_error is not None:
            if self.api.session.pass_error:
                return self.empty_objects_error().error
            else:
                raise self.empty_objects_error
        else:
            return result


scraper = APIScraperMethod
multiscraper = APIScraperMultiMethod


def with_init(coro):
    @wraps(coro)
    async def wrapper(self: scraper, **kwargs):
        if not self.api.session.cookies:
            raise CookieError('Cookie jar is empty. Set cookies.')
        init_result = await self.init(**kwargs)
        if isinstance(init_result, dict):
            return init_result
        else:
            return await coro(self, **kwargs)

    return wrapper


class GroupsGet(scraper):
    """Returns a list of the communities to which the current user belongs."""

    url = 'https://my.mail.ru/my/communities'

    class Scripts(scraper.s):
        class Selectors(scraper.ss):
            groups = (
                f'{scraper.ss.content}'
                ' div.groups-catalog'
                ' div.groups-catalog__mine-groups'
                ' div.groups-catalog__small-groups'
            )
            bar = f'{groups} div.groups-catalog__groups-more'
            hidden_bar = scraper.ssst.hidden % bar
            visible_bar = scraper.ssst.visible % bar
            catalog = f'{groups} div.groups__container'
            button = f'{bar} span.ui-button-gray'
            progress_button = f'{bar} span.progress'
            item = f'{catalog} div.groups__item'

        click = scraper.sst.click % Selectors.button
        button_class = scraper.sst.getattr % 'class'
        bar_css = scraper.sst.getattr % 'style'
        loaded = f'{scraper.sst.length % Selectors.item} > %d'

    s = Scripts
    ss = Scripts.Selectors

    async def init(self, limit=10, offset=0, ext=0):
        info = await self.api.users.getInfo(uids='')
        if isinstance(info, dict):
            return info
        url = self.url
        log.debug(f'go to {url}')
        self.page = await self.api.page(
            url,
            self.api.session.session_key,
            self.api.session.cookies,
        )
        _ = await self.page.screenshot()
        return True

    @with_init
    async def call(self, *, limit=10, offset=0, ext=0):
        return await self.scrape(ext, limit, offset)

    async def scrape(self, ext, limit, offset):
        """Appends groups from the `page` to the `groups` list."""
        log.debug(f'scrape subset: offset={offset}, limit={limit}')

        groups, cnt = [], 0
        async for group in self.groups(ext):
            if cnt < offset:
                continue
            else:
                groups.append(group)
            cnt += 1

            if len(groups) >= limit:
                break

        return groups

    async def groups(self, ext):
        hidden_bar, bar, elements = None, True, []

        while True:
            offset = len(elements)
            catalog = await self.page.J(self.ss.catalog)
            elements = await catalog.JJ(self.ss.item)
            for element in elements[offset:]:
                item = await GroupItem.from_element(element)
                link = item['link'].lstrip('/')
                resp = await self.api.session.public_request([link])
                group = await self.api.users.getInfo(uids=resp['uid'])
                yield group[0] if ext else group[0]['uid']

            if await self.page.J(self.ss.button):
                await self.page.evaluate(self.s.click)

            progress_button = True
            while progress_button:
                progress_button = await self.page.J(self.ss.progress_button)

            # if current iteration should be the last
            if hidden_bar is not None or bar is None:
                break

            # last iteration flag
            bar = await self.page.J(self.ss.bar)  # if is absent
            hidden_bar = await self.page.J(self.ss.hidden_bar)  # if is present


class GroupsGetInfo(multiscraper):
    """Returns information about communities by their IDs."""

    class Scripts(multiscraper.s):
        class Selectors(multiscraper.ss):
            pass

    s = Scripts
    ss = Scripts.Selectors

    empty_objects_error = EmptyGroupsError
    ignored_errors = (APIError, KeyError)  # KeyError when group_info is absent

    async def init(self, uids=''):
        info = await self.api.users.getInfo(uids=uids)
        if isinstance(info, dict):
            return info
        url = info[0]['link']
        log.debug(f'go to {url}')
        self.page = await self.api.page(
            url,
            self.api.session.session_key,
            self.api.session.cookies,
            True,
        )
        _ = await self.page.screenshot()
        return True

    @with_init
    async def call(self, *, uids=''):
        return await self.scrape(uids)

    async def scrape(self, uids):
        """Returns additional information about a group.

        Object fields that are scraped here:
            - 'is_closed' - information whether the group's stream events
                are closed for current user.

        """

        info = await self.api.users.getInfo(uids=uids)
        signage = await self.page.J(self.ss.closed_signage)
        is_closed = True if signage is not None else False
        info[0]['group_info'].update({'is_closed': is_closed})
        return info


class GroupsJoin(scraper):
    """With this method you can join the group."""

    retry_interval = 1
    num_attempts = 10

    class Scripts(scraper.s):
        class Selectors(scraper.ss):
            links = (
                f'{scraper.ss.profile_content}'
                ' div.profile__activeLinks'
                ' div.profile__activeLinks_community'
            )
            join_span = f'{links} span.profile__activeLinks_button_enter'
            sent_span = f'{links} span.profile__activeLinks_link_modarated'
            approved_span = f'{links} span.profile__activeLinks_link_inGroup'
            auth_span = f'{links} div.l-popup_community-authorization'

        join_span_visible = scraper.sst.visible % Selectors.join_span
        sent_span_visible = scraper.sst.visible % Selectors.sent_span
        approved_span_visible = scraper.sst.visible % Selectors.approved_span

        join_click = f'{scraper.sst.click % Selectors.join_span}'

    s = Scripts
    ss = Scripts.Selectors

    async def init(self, group_id=''):
        info = await self.api.users.getInfo(uids=group_id)
        if isinstance(info, dict):
            return info
        url = info[0]['link']
        log.debug(f'go to {url}')
        self.page = await self.api.page(
            url,
            self.api.session.session_key,
            self.api.session.cookies,
            True,
        )
        _ = await self.page.screenshot()
        return True

    @with_init
    async def call(self, *, group_id=''):
        return await self.scrape()

    async def scrape(self):
        if await self.page.evaluate(self.s.join_span_visible):
            return await self.join()
        elif await self.page.evaluate(self.s.sent_span_visible):
            return 1
        elif await self.page.evaluate(self.s.approved_span_visible):
            return 1
        else:
            raise APIScrapperError('A join button not found.')

    async def join(self):
        for i in range(self.num_attempts):
            await self.page.evaluate(self.s.join_click)
            await asyncio.sleep(self.retry_interval)

            if await self.page.evaluate(self.s.sent_span_visible):
                return 1
            elif await self.page.evaluate(self.s.approved_span_visible):
                return 1

        raise APIScrapperError('Failed to send join request.')


class StreamGetByAuthor(scraper):
    """Returns a list of events from user or community stream by their IDs."""

    class Scripts(scraper.s):
        class Selectors(scraper.ss):
            feed = f'{scraper.ss.main_page} div.b-community__main-page__feed'
            stream = f'{feed} div.b-history[data-state]'
            updating_stream = f'{stream}[data-state=""]'
            loading_stream = f'{stream}[data-state="loading"]'
            loaded_stream = f'{stream}[data-state="loaded"]'
            ended_stream = f'{stream}[data-state="noevents"]'
            content = f'{feed} div.content-wrapper'
            event = f'{stream} div.b-history-event[data-astat]'

        stream_state = scraper.sst.getattr % 'data-state'

    s = Scripts
    ss = Scripts.Selectors

    async def init(self, uid='', limit=10, skip=''):
        info = await self.api.users.getInfo(uids=uid)
        if isinstance(info, dict):
            return info
        url = info[0]['link']
        log.debug(f'go to {url}')
        self.page = await self.api.page(
            url,
            self.api.session.session_key,
            self.api.session.cookies,
        )
        _ = await self.page.screenshot()
        return True

    @with_init
    async def call(self, *, uid='', limit=10, skip=''):
        return await self.scrape(limit, skip)

    async def scrape(self, limit, skip):
        """Returns a list of events from user or community stream."""

        log.debug(f'scrape subset: skip={skip}, limit={limit}')

        try:
            events = []
            async for event in self.stream():
                if skip:
                    skip = skip if event['id'] != skip else False
                else:
                    events.append(event)

                if len(events) >= limit:
                    break
        except (AccessDeniedError, BlackListError) as e:
            if self.api.session.pass_error:
                return e.error
            else:
                raise e

        return events

    async def stream(self):
        """Yields stream events from the beginning to the end."""

        ended_stream, elements = None, []

        while True:
            offset = len(elements)
            content = await self.page.J(self.ss.content)

            if content is None:
                signage = await self.page.J(self.ss.closed_signage)
                if signage:
                    raise AccessDeniedError()
                else:
                    raise BlackListError()

            elements = await content.JJ(self.ss.event)
            for element in elements[offset:]:
                yield await Event.from_element(element)

            await self.page.evaluate(self.s.scroll)

            loading_stream, updating_stream = True, True
            while loading_stream or updating_stream:
                stream = False
                while not stream and content:
                    stream = await self.page.waitForSelector(self.ss.stream)
                    content = await self.page.J(self.ss.content)

                loading_stream = await self.page.J(self.ss.loading_stream)
                updating_stream = await self.page.J(self.ss.updating_stream)

            # if current iteration should be the last
            if ended_stream is not None:
                break

            # last iteration flag
            ended_stream = await self.page.J(self.ss.ended_stream)


scrapers = {
    'groups': APIScraperMethod,
    'groups.get': GroupsGet,
    'groups.getInfo': GroupsGetInfo,
    'groups.join': GroupsJoin,
    'stream': APIScraperMethod,
    'stream.getByAuthor': StreamGetByAuthor,
}



import logging.config
import yaml
from os.path import dirname, join


with open(join(dirname(__file__), 'config.yaml')) as f:
    config = yaml.safe_load(f)


logging.config.dictConfig(config['logging'])
